@@ -1,4 +1,4 @@
-# Copyright 2014 Google Inc.
+# Copyright 2014 Google LLC
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -18,25 +18,20 @@ from google.rpc import status_pb2
 
 from google.cloud import _http as connection_module
 from google.cloud import exceptions
-from google.cloud.proto.datastore.v1 import datastore_pb2 as _datastore_pb2
-
-from google.cloud.datastore import __version__
+from google.cloud.datastore_v1.proto import datastore_pb2 as _datastore_pb2
 
 
-DATASTORE_API_HOST = 'datastore.googleapis.com'
+DATASTORE_API_HOST = "datastore.googleapis.com"
 """Datastore API request host."""
-API_BASE_URL = 'https://' + DATASTORE_API_HOST
+API_BASE_URL = "https://" + DATASTORE_API_HOST
 """The base of the API call URL."""
-API_VERSION = 'v1'
+API_VERSION = "v1"
 """The version of the API, used in building the API call's URL."""
-API_URL_TEMPLATE = ('{api_base}/{api_version}/projects'
-                    '/{project}:{method}')
+API_URL_TEMPLATE = "{api_base}/{api_version}/projects" "/{project}:{method}"
 """A template for the URL of a particular API call."""
 
-_CLIENT_INFO = connection_module.CLIENT_INFO_TEMPLATE.format(__version__)
 
-
-def _request(http, project, method, data, base_url):
+def _request(http, project, method, data, base_url, client_info):
     """Make a request over the Http transport to the Cloud Datastore API.
 
     :type http: :class:`requests.Session`
@@ -56,30 +51,34 @@ def _request(http, project, method, data, base_url):
     :type base_url: str
     :param base_url: The base URL where the API lives.
 
+    :type client_info: :class:`google.api_core.client_info.ClientInfo`
+    :param client_info: used to generate user agent.
+
     :rtype: str
     :returns: The string response content from the API call.
     :raises: :class:`google.cloud.exceptions.GoogleCloudError` if the
              response code is not 200 OK.
     """
+    user_agent = client_info.to_user_agent()
     headers = {
-        'Content-Type': 'application/x-protobuf',
-        'User-Agent': connection_module.DEFAULT_USER_AGENT,
-        connection_module.CLIENT_INFO_HEADER: _CLIENT_INFO,
+        "Content-Type": "application/x-protobuf",
+        "User-Agent": user_agent,
+        connection_module.CLIENT_INFO_HEADER: user_agent,
     }
     api_url = build_api_url(project, method, base_url)
 
-    response = http.request(
-        url=api_url, method='POST', headers=headers, data=data)
+    response = http.request(url=api_url, method="POST", headers=headers, data=data)
 
     if response.status_code != 200:
         error_status = status_pb2.Status.FromString(response.content)
         raise exceptions.from_http_status(
-            response.status_code, error_status.message, errors=[error_status])
+            response.status_code, error_status.message, errors=[error_status]
+        )
 
     return response.content
 
 
-def _rpc(http, project, method, base_url, request_pb, response_pb_cls):
+def _rpc(http, project, method, base_url, client_info, request_pb, response_pb_cls):
     """Make a protobuf RPC request.
 
     :type http: :class:`requests.Session`
@@ -95,6 +94,9 @@ def _rpc(http, project, method, base_url, request_pb, response_pb_cls):
     :type base_url: str
     :param base_url: The base URL where the API lives.
 
+    :type client_info: :class:`google.api_core.client_info.ClientInfo`
+    :param client_info: used to generate user agent.
+
     :type request_pb: :class:`google.protobuf.message.Message` instance
     :param request_pb: the protobuf instance representing the request.
 
@@ -107,8 +109,7 @@ def _rpc(http, project, method, base_url, request_pb, response_pb_cls):
     :returns: The RPC message parsed from the response.
     """
     req_data = request_pb.SerializeToString()
-    response = _request(
-        http, project, method, req_data, base_url)
+    response = _request(http, project, method, req_data, base_url, client_info)
     return response_pb_cls.FromString(response)
 
 
@@ -132,8 +133,8 @@ def build_api_url(project, method, base_url):
     :returns: The API URL created.
     """
     return API_URL_TEMPLATE.format(
-        api_base=base_url, api_version=API_VERSION,
-        project=project, method=method)
+        api_base=base_url, api_version=API_VERSION, project=project, method=method
+    )
 
 
 class HTTPDatastoreAPI(object):
@@ -148,48 +149,52 @@ class HTTPDatastoreAPI(object):
     def __init__(self, client):
         self.client = client
 
-    def lookup(self, project, read_options, key_pbs):
+    def lookup(self, project_id, keys, read_options=None):
         """Perform a ``lookup`` request.
 
-        :type project: str
-        :param project: The project to connect to. This is
-                        usually your project name in the cloud console.
+        :type project_id: str
+        :param project_id: The project to connect to. This is
+                           usually your project name in the cloud console.
+
+        :type keys: List[.entity_pb2.Key]
+        :param keys: The keys to retrieve from the datastore.
 
         :type read_options: :class:`.datastore_pb2.ReadOptions`
-        :param read_options: The options for this lookup. Contains a
+        :param read_options: (Optional) The options for this lookup. Contains
                              either the transaction for the read or
                              ``STRONG`` or ``EVENTUAL`` read consistency.
 
-        :type key_pbs: list of
-                       :class:`.entity_pb2.Key`
-        :param key_pbs: The keys to retrieve from the datastore.
-
         :rtype: :class:`.datastore_pb2.LookupResponse`
         :returns: The returned protobuf response object.
         """
         request_pb = _datastore_pb2.LookupRequest(
-            project_id=project,
-            read_options=read_options,
-            keys=key_pbs,
+            project_id=project_id, read_options=read_options, keys=keys
+        )
+        return _rpc(
+            self.client._http,
+            project_id,
+            "lookup",
+            self.client._base_url,
+            self.client._client_info,
+            request_pb,
+            _datastore_pb2.LookupResponse,
         )
-        return _rpc(self.client._http, project, 'lookup',
-                    self.client._base_url,
-                    request_pb, _datastore_pb2.LookupResponse)
 
-    def run_query(self, project, partition_id, read_options,
-                  query=None, gql_query=None):
+    def run_query(
+        self, project_id, partition_id, read_options=None, query=None, gql_query=None
+    ):
         """Perform a ``runQuery`` request.
 
-        :type project: str
-        :param project: The project to connect to. This is
-                        usually your project name in the cloud console.
+        :type project_id: str
+        :param project_id: The project to connect to. This is
+                           usually your project name in the cloud console.
 
         :type partition_id: :class:`.entity_pb2.PartitionId`
         :param partition_id: Partition ID corresponding to an optional
                              namespace and project ID.
 
         :type read_options: :class:`.datastore_pb2.ReadOptions`
-        :param read_options: The options for this query. Contains a
+        :param read_options: (Optional) The options for this query. Contains
                              either the transaction for the read or
                              ``STRONG`` or ``EVENTUAL`` read consistency.
 
@@ -205,37 +210,52 @@ class HTTPDatastoreAPI(object):
         :returns: The returned protobuf response object.
         """
         request_pb = _datastore_pb2.RunQueryRequest(
-            project_id=project,
+            project_id=project_id,
             partition_id=partition_id,
             read_options=read_options,
             query=query,
             gql_query=gql_query,
         )
-        return _rpc(self.client._http, project, 'runQuery',
-                    self.client._base_url,
-                    request_pb, _datastore_pb2.RunQueryResponse)
+        return _rpc(
+            self.client._http,
+            project_id,
+            "runQuery",
+            self.client._base_url,
+            self.client._client_info,
+            request_pb,
+            _datastore_pb2.RunQueryResponse,
+        )
 
-    def begin_transaction(self, project):
+    def begin_transaction(self, project_id, transaction_options=None):
         """Perform a ``beginTransaction`` request.
 
-        :type project: str
-        :param project: The project to connect to. This is
-                        usually your project name in the cloud console.
+        :type project_id: str
+        :param project_id: The project to connect to. This is
+                           usually your project name in the cloud console.
+
+        :type transaction_options: ~.datastore_v1.types.TransactionOptions
+        :param transaction_options: (Optional) Options for a new transaction.
 
         :rtype: :class:`.datastore_pb2.BeginTransactionResponse`
         :returns: The returned protobuf response object.
         """
         request_pb = _datastore_pb2.BeginTransactionRequest()
-        return _rpc(self.client._http, project, 'beginTransaction',
-                    self.client._base_url,
-                    request_pb, _datastore_pb2.BeginTransactionResponse)
+        return _rpc(
+            self.client._http,
+            project_id,
+            "beginTransaction",
+            self.client._base_url,
+            self.client._client_info,
+            request_pb,
+            _datastore_pb2.BeginTransactionResponse,
+        )
 
-    def commit(self, project, mode, mutations, transaction=None):
+    def commit(self, project_id, mode, mutations, transaction=None):
         """Perform a ``commit`` request.
 
-        :type project: str
-        :param project: The project to connect to. This is
-                        usually your project name in the cloud console.
+        :type project_id: str
+        :param project_id: The project to connect to. This is
+                           usually your project name in the cloud console.
 
         :type mode: :class:`.gapic.datastore.v1.enums.CommitRequest.Mode`
         :param mode: The type of commit to perform. Expected to be one of
@@ -254,51 +274,68 @@ class HTTPDatastoreAPI(object):
         :returns: The returned protobuf response object.
         """
         request_pb = _datastore_pb2.CommitRequest(
-            project_id=project,
+            project_id=project_id,
             mode=mode,
             transaction=transaction,
             mutations=mutations,
         )
-        return _rpc(self.client._http, project, 'commit',
-                    self.client._base_url,
-                    request_pb, _datastore_pb2.CommitResponse)
+        return _rpc(
+            self.client._http,
+            project_id,
+            "commit",
+            self.client._base_url,
+            self.client._client_info,
+            request_pb,
+            _datastore_pb2.CommitResponse,
+        )
 
-    def rollback(self, project, transaction_id):
+    def rollback(self, project_id, transaction):
         """Perform a ``rollback`` request.
 
-        :type project: str
-        :param project: The project to connect to. This is
-                        usually your project name in the cloud console.
+        :type project_id: str
+        :param project_id: The project to connect to. This is
+                           usually your project name in the cloud console.
 
-        :type transaction_id: bytes
-        :param transaction_id: The transaction ID to rollback.
+        :type transaction: bytes
+        :param transaction: The transaction ID to rollback.
 
         :rtype: :class:`.datastore_pb2.RollbackResponse`
         :returns: The returned protobuf response object.
         """
         request_pb = _datastore_pb2.RollbackRequest(
-            project_id=project,
-            transaction=transaction_id,
+            project_id=project_id, transaction=transaction
         )
         # Response is empty (i.e. no fields) but we return it anyway.
-        return _rpc(self.client._http, project, 'rollback',
-                    self.client._base_url,
-                    request_pb, _datastore_pb2.RollbackResponse)
+        return _rpc(
+            self.client._http,
+            project_id,
+            "rollback",
+            self.client._base_url,
+            self.client._client_info,
+            request_pb,
+            _datastore_pb2.RollbackResponse,
+        )
 
-    def allocate_ids(self, project, key_pbs):
+    def allocate_ids(self, project_id, keys):
         """Perform an ``allocateIds`` request.
 
-        :type project: str
-        :param project: The project to connect to. This is
-                        usually your project name in the cloud console.
+        :type project_id: str
+        :param project_id: The project to connect to. This is
+                           usually your project name in the cloud console.
 
-        :type key_pbs: list of :class:`.entity_pb2.Key`
-        :param key_pbs: The keys for which the backend should allocate IDs.
+        :type keys: List[.entity_pb2.Key]
+        :param keys: The keys for which the backend should allocate IDs.
 
         :rtype: :class:`.datastore_pb2.AllocateIdsResponse`
         :returns: The returned protobuf response object.
         """
-        request_pb = _datastore_pb2.AllocateIdsRequest(keys=key_pbs)
-        return _rpc(self.client._http, project, 'allocateIds',
-                    self.client._base_url,
-                    request_pb, _datastore_pb2.AllocateIdsResponse)
+        request_pb = _datastore_pb2.AllocateIdsRequest(keys=keys)
+        return _rpc(
+            self.client._http,
+            project_id,
+            "allocateIds",
+            self.client._base_url,
+            self.client._client_info,
+            request_pb,
+            _datastore_pb2.AllocateIdsResponse,
+        )
