@@ -239,7 +239,10 @@ class Client(ClientWithProject):
         _use_grpc=None,
     ):
         super(Client, self).__init__(
-            project=project, credentials=credentials, _http=_http
+            project=project,
+            credentials=credentials,
+            client_options=client_options,
+            _http=_http,
         )
         self.namespace = namespace
         self._client_info = client_info
@@ -667,7 +670,7 @@ class Client(ClientWithProject):
         """Reserve a list of IDs from a complete key.
 
         :type complete_key: :class:`google.cloud.datastore.key.Key`
-        :param complete_key: Partial key to use as base for reserved IDs.
+        :param complete_key: Complete key to use as base for reserved IDs.
 
         :type num_ids: int
         :param num_ids: The number of IDs to reserve.
